/**
* Represents a parameter, fixed, environmental, or genotypical
* that can be encoded in a chromosome
* The value is scaled to fit in positive integer space.  Modified
* from original so each param is represented as 4 character hex value
* in encoding.
* This code was originally written by Ariel Dolan
* @author Groff 1/2003
*/
class EfloyParam {

        float min;  // minimum value for this param
        float max;  // maximum "
        float step; // scale factor
        float value;// param value
        int nsteps; // range
        String name;// param name
        boolean mutatable; // mutable

	public EfloyParam(float mn, float mx, float st, float vl, String nm, boolean mt) {

		min = mn;
		max = mx;
		step = st;
		name = nm;
		value = vl;
		mutatable = mt;

		nsteps = (int) ((mx-mn)/step);
		//Efloys.deb.showMsg("Given step. name = "+nm+" mx= "+mx+" mn = "+mn+" step= "+step+" NSTEPS= "+nsteps);


	}

	public EfloyParam (float mn, float mx, int ns, float vl, String nm, boolean mt) {

		min = mn;
		max = mx;
		nsteps = ns;
		name = nm;
		value = vl;
		mutatable = mt;

		step = (float) ((mx-mn)/nsteps);
		//Efloys.deb.showMsg("Given ns. name = "+nm+" mx= "+mx+" mn = "+mn+" nsteps= "+nsteps+" STEP= "+step);

	}
        /*
	char EncodeValue() {

		int n = (int) ((value - min)/step);
		char k = (char) (65+n);
		return k;

	}

	float DecodeValue(char kar) {

		int n;
		float k;

		try {
			n = (int) (kar - 65);
			k = min + n*step;
			return k;
		}
		catch (Exception e) {
			//Efloys.deb.showMsg("DecodeValue. kar= "+kar+" n= "+n+" step = "+step+" min = "+min+" k= "+k);
			//Efloys.deb.showMsg("DecodeValue. kar= "+kar+" e = "+e.toString());
			return 0f;
		}

	}
        */

        String EncodeValue() {
                int n = (int) ((value - min)/step);
                //System.out.println("Encode "+n);
                String er = Long.toHexString((long)n).toUpperCase();
                if( er.length() < 4 ) er = "0000".substring(0,4-er.length())+er;
                if( er.length() > 4 ) er = er.substring(er.length()-4);
                //System.out.println("Encoded "+er);
                return er;
	}

        float DecodeValue(String kar) {
                float k;
		try {
                        long val = Long.valueOf(kar, 16).longValue();
                        k = min + val * step;
                        return k;
		}
		catch (Exception e) {
			//Efloys.deb.showMsg("DecodeValue. kar= "+kar+" n= "+n+" step = "+step+" min = "+min+" k= "+k);
			//Efloys.deb.showMsg("DecodeValue. kar= "+kar+" e = "+e.toString());
			return 0f;
		}

	}
}

