/**
* Greater Object Oriented eFloys - GOOeFloys
* This module is original, derived from GooEfloy it represents a predator
* @author J. Groff 1/2003
*/
import java.awt.*;

public class PredEfloy extends GooEfloy {

        /**
        * Construct this predator.  The evolutionary ladder factor is 1
        */
        public PredEfloy(EfloyParam pr[], String cr, EfloyParam fp[], String fx,
		EfloyParam ep[], String ev) {
			super(pr, cr, fp, fx, ep, ev);
                        eLadderFactor = 1;
                        color = Color.red;
                        //initialEnergy = LifeSpan * 100;
                        initialEnergy *= 2;
	}

        /**
        * Collision with same species. Increase cooperation
        * @param collider The floy thats collided with us
        * @return The collision factor for movement
        */
        private int collideSameSpecies(GooEfloy collider) {
                // feeding function
                //if( collider.mother == id ) {
                //        --MaxEnergyDose;
                //        Efloys.deb.showMsg("Mom "+id+" fed "+collider.id);
                //} else {
                // under the right circumstances, could a surrogate mommy form?
                //        if( mother == collider.id ) {
                //                ++MaxEnergyDose;
                //                Efloys.deb.showMsg("Kid "+id+" fed by "+collider.id);
                //        }
                //}
                //safety += (int) MaxSafetyDose;
                // lets make collision a lack of cooperation for a hunter
                cooperation -= MaxCooperationDose;
                if( cooperation < 0 ) cooperation = 0;
		return (int) CollisionBrotherFactor;
	}
	
        private int approachSameSpecies(int neighborDist) {
                int CoopAddition;
                CoopAddition = (int) ((MaxCooperationDose*CollisionDistance)/(neighborDist+CollisionDistance));
                cooperation += CoopAddition;
		return (int) DistBrotherFactor;
	}

        private int approachDifferentSpecies() {
             return (int) DistStrangerFactor;
        }

	/**
	* Here we can code predator/prey behavior in subclasses
        * a PredEFloy is predator because of behavior in this and related methods
        * @param neighborStrange The collider
        * @return The movement mod factor
	*/
        private int collideDifferentSpecies(GooEfloy neighborStrange) {
		int rev;
                //Efloys.deb.showMsg(this.getClass()+" "+neighborStrange.getClass());
                // Metabolic rate 20% (they lose 10x max dos, we get 2x)
                energy += (int) MaxEnergyDose * 2;
                rev = (int) CollisionStrangerFactor;
                neighborStrange.energy -= MaxEnergyDose*10.0F;
                neighborStrange.safety -= MaxSafetyDose*10.0F; 
                if (neighborStrange.energy < 0) neighborStrange.energy = 0;
                if (neighborStrange.safety < 0) neighborStrange.safety = 0;
                neighborStrange.GetFitness();
                        if (Efloys.CurrentGeneration > 0) 
                                Efloys.appcontext.showStatus("Generation #" + Efloys.CurrentGeneration+"  predator "+id+" energy: "+energy);
                        else
                                Efloys.appcontext.showStatus("Preditor "+id+" energy: "+energy);
                //Efloys.deb.showMsg(id+" "+neighborStrange.id+" "+energy+" "+neighborStrange.energy);
		return rev;
	}
	/**
        * Process each neighbor of this GOOeFloy
	*/
	public void Process() {
		int rev;
		int i;
		int d;

		xtail = x;
		ytail = y;
		rev = -1;
                //Efloys.deb.showMsg("Process predator "+id);
		for (i=0;i<NumberOfNeighbors;i++) {
			
			d = dist(neighbors[i]);

			if (d == 0)
				rev = 0;
			else 
				if (d < CollisionDistance) {
					if( !neighbors[i].getClass().equals(this.getClass()) ) {
                                                if( d < 10 )
                                                        rev = collideDifferentSpecies(neighbors[i]);
                                                else
                                                        rev = approachDifferentSpecies();
					} else { 
                                                rev = collideSameSpecies(neighbors[i]);
					}
				} else  {
					if( !neighbors[i].getClass().equals(this.getClass()) ) {
                                                rev = approachDifferentSpecies();
                                        } else {
                                                rev = approachSameSpecies(d);
                                        }
				}

				if (x < neighbors[i].x)
					vx += ApproachAcceleration*rev;
				else
					vx -= ApproachAcceleration*rev;

				if (y < neighbors[i].y)
					vy += ApproachAcceleration*rev;
				else
					vy -= ApproachAcceleration*rev;
		}
	

		if (vx > MaxSpeed) vx = MaxSpeed;
		if (vx < -MaxSpeed) vx = -MaxSpeed;
		if (vy > MaxSpeed) vy = MaxSpeed;
		if (vy < -MaxSpeed) vy = -MaxSpeed;

		if (x < 1) vx = BounceSpeed;
		if (x > (width-3)) vx = -BounceSpeed;
		//if (xtail > (width-3)) vx = -BounceSpeed;
		if (y < 1) vy = BounceSpeed;
		if (y > (height-1)) vy = -BounceSpeed;
		
                float v = (float)(Math.sqrt(vx*vx + vy*vy));
                energy -= (int)v;

                //Efloys.deb.showMsg("Predator "+v+" "+MaxSpeed+" "+MaxEnergyDose);
                        if (x < width/2) vx += CenterAcceleration;
                        if (x > width/2) vx -= CenterAcceleration;
                        if (y < height/2) vy += CenterAcceleration;
                        if (y > height/2) vy -= CenterAcceleration;

		x += vx;
		y += vy;
		if (energy < 1)	{
			if (Efloys.WithSound)
                                Efloys.preddie.play();
                        energy = initialEnergy;
                        GetFitness();
                        if( Efloys.InEvolution ) {
                                // is this the last predator to die?
                                int ipred = 0;
                                for(; ipred < Efloys.Efloys.length; ipred++)
                                        if( Efloys.Efloys[ipred].getClass().equals(this.getClass()) &&
                                        !Efloys.Efloys[ipred].equals(this) ) break;
                                // If not, this dead one rots and breeds lower forms
                                // If in evolutionary mode.
                                if( ipred == Efloys.Efloys.length) {
                                        Efloys.NewGeneration = true;
                                        //Efloys.deb.showMsg("New gen flag set");
                                } else {
                                        Efloys.instance.Rot(this);
                                }
                        } else {
                                x = xtail = Efloys.instance.canvas.size().width;
                                y = ytail = 0;
                        }
                        return;
		}

		GetFitness();

	}

        /**
        * Get the display type name, implement as you see fit in subclasses
        * @return The string display name for this type
        */
	public String GetTypeName() {
                return getClass().getName();
	}

        /**
        * Make a baby via factory; result is same type as subclass.
        * For predator, we have the kid born on top of mom
        * @param pr Param array
        * @param cr Crossover of Mom and Dads chromosomes
        * @param fp Fixed param array
        * @param fx Fixed param string ?
        * @param ep Environmental params array
        * @param ev Enviro params string ?
        * @return The child as GooEfloy
        */
        public GooEfloy BabyFactory(EfloyParam pr[], String cr, EfloyParam fp[], String fx,
                EfloyParam ep[], String ev) {
                GooEfloy g = (GooEfloy)(new PredEfloy(pr, cr, fp, fx, ep, ev));
                g.mother = id;
                g.x = x;
                g.y = y;
                GetNeighbors();
                return g;
        }

        public void SmoothingFunction(int coopFactor, int safeFactor) {
                cooperation -= coopFactor;
                if( cooperation < 0 ) cooperation = 0;
                safety -= safeFactor;
                if( safety < 0 ) safety = 0;
        }
        /**
        * Determines if we ignore a certain species for some purpose, such as
        * moving toward or away from the...
        * @param ignore The GOOeFloy to possibly ignore if it meets criteria we set
        * @return true or false to ignore
        */
        public boolean IgnoreSpecies(GooEfloy ignore) {
                if( this.getClass().equals(ignore.getClass()) ) return true;
                return false;
        }
}
